@@ -17,14 +17,17 @@ from wand.api import library
 from wand.color import Color
 from wand.image import Image
 
-from pcbnewTransition import pcbnew, pcbnewVersion, isV6
-from lxml import etree
+from pcbnewTransition import pcbnew, KICAD_VERSION, isV6
+from lxml import etree, objectify
+
 
 # Give more priority to local modules than installed versions
 PKG_BASE = os.path.dirname(__file__)
 sys.path.insert(0, os.path.dirname(os.path.abspath(PKG_BASE)))
 from pcbdraw import __version__
 
+etree.register_namespace("xlink", "http://www.w3.org/1999/xlink")
+
 STYLES_SUBDIR = 'styles'
 FOOTPRINTS_SUBDIR = 'footprints'
 data_path = [PKG_BASE]
@@ -209,8 +212,8 @@ def mm2ki(val):
     return val * 1000000
 
 # KiCAD 5 and KiCAD 6 use different units of the SVG
-ki2svg = (lambda x: x) if isV6(pcbnewVersion) else ki2dmil
-svg2ki = (lambda x: x) if isV6(pcbnewVersion) else dmil2ki
+ki2svg = (lambda x: x) if isV6(KICAD_VERSION) else ki2dmil
+svg2ki = (lambda x: x) if isV6(KICAD_VERSION) else dmil2ki
 
 def to_kicad_basic_units(val):
     """
@@ -249,6 +252,14 @@ def to_user_units(val):
     if unit == "in":
         return 90
 
+def make_XML_identifier(s):
+    """
+    Given a name, strip invalid characters from XML identifier
+    """
+    s = re.sub('[^0-9a-zA-Z_]', '', s)
+    s = re.sub('^[^a-zA-Z_]+', '', s)
+    return s
+
 def read_svg_unique(filename):
     prefix = unique_prefix() + "_"
     root = etree.parse(filename).getroot()
@@ -270,7 +281,7 @@ def read_svg_unique(filename):
     return root
 
 def extract_svg_content(root):
-    # Remove SVG namespace to ease our lifes and change ids
+    # Remove SVG namespace to ease our lives and change ids
     for el in root.getiterator():
         if '}' in str(el.tag):
             el.tag = el.tag.split('}', 1)[1]
@@ -347,7 +358,7 @@ def get_board_polygon(svg_elements):
     e = etree.Element("path", d=path, style="fill-rule: evenodd;")
     return e
 
-def process_board_substrate_layer(container, name, source, colors, boardsize):
+def process_board_substrate_layer(container, defs, name, source, colors, boardsize):
     layer = etree.SubElement(container, "g", id="substrate-" + name,
         style="fill:{0}; stroke:{0};".format(colors[name]))
     if name == "pads":
@@ -358,8 +369,8 @@ def process_board_substrate_layer(container, name, source, colors, boardsize):
         strip_fill_svg(element)
         layer.append(element)
 
-def process_board_substrate_base(container, name, source, colors, boardsize):
-    clipPath = etree.SubElement(etree.SubElement(container, "defs"), "clipPath")
+def process_board_substrate_base(container, defs, name, source, colors, boardsize):
+    clipPath = etree.SubElement(defs, "clipPath")
     clipPath.attrib["id"] = "cut-off"
     clipPath.append(get_board_polygon(extract_svg_content(read_svg_unique(source))))
 
@@ -372,8 +383,8 @@ def process_board_substrate_base(container, name, source, colors, boardsize):
         strip_fill_svg(element)
         outline.append(element)
 
-def process_board_substrate_mask(container, name, source, colors, boardsize):
-    mask = etree.SubElement(etree.SubElement(container, "defs"), "mask")
+def process_board_substrate_mask(container, defs, name, source, colors, boardsize):
+    mask = etree.SubElement(defs, "mask")
     mask.attrib["id"] = name
     for element in extract_svg_content(read_svg_unique(source)):
         for item in element.getiterator():
@@ -381,7 +392,7 @@ def process_board_substrate_mask(container, name, source, colors, boardsize):
                 # KiCAD plots in black, for mask we need white
                 item.attrib["style"] = item.attrib["style"].replace("#000000", "#ffffff")
         mask.append(element)
-    silkMask = etree.SubElement(etree.SubElement(container, "defs"), "mask")
+    silkMask = etree.SubElement(defs, "mask")
     silkMask.attrib["id"] = name + "-silkscreen"
     bg = etree.SubElement(silkMask, "rect", attrib={
         "x": str(ki2svg(boardsize.GetX())),
@@ -394,7 +405,7 @@ def process_board_substrate_mask(container, name, source, colors, boardsize):
         # KiCAD plots black, no need to change fill
         silkMask.append(element)
 
-def get_layers(board, colors, toPlot):
+def get_layers(board, colors, defs, toPlot):
     """
     Plot given layers, process them and return them as <g>
     """
@@ -419,14 +430,15 @@ def get_layers(board, colors, toPlot):
                 pctl.SetLayer(l)
                 pctl.PlotLayer()
         pctl.ClosePlot()
+        board.SetVisibleAlls()
         boardsize = board.ComputeBoundingBox()
         for f, _, process in toPlot:
             for svg_file in os.listdir(tmp):
                 if svg_file.endswith("-" + f + ".svg"):
-                    process(container, f, os.path.join(tmp, svg_file), colors, boardsize)
+                    process(container, defs, f, os.path.join(tmp, svg_file), colors, boardsize)
     return container
 
-def get_board_substrate(board, colors, holes, back):
+def get_board_substrate(board, colors, defs, holes, back):
     """
     Plots all front layers from the board and arranges them in a visually appealing style.
     return SVG g element with the board substrate
@@ -473,14 +485,15 @@ def get_board_substrate(board, colors, holes, back):
                 pctl.SetLayer(l)
                 pctl.PlotLayer()
         pctl.ClosePlot()
+        board.SetVisibleAlls()
         boardsize = board.ComputeBoundingBox()
         for f, _, process in toPlot:
             for svg_file in os.listdir(tmp):
                 if svg_file.endswith("-" + f + ".svg"):
-                    process(container, f, os.path.join(tmp, svg_file), colors, boardsize)
+                    process(container, defs, f, os.path.join(tmp, svg_file), colors, boardsize)
 
     if holes:
-        container.append(get_hole_mask(board))
+        get_hole_mask(board, defs)
         container.attrib["mask"] = "url(#hole-mask)"
     return container
 
@@ -503,11 +516,11 @@ def walk_components(board, back, export):
         pos = (center.x, center.y, orient)
         export(lib, name, value, ref, pos)
 
-def get_hole_mask(board):
-    defs = etree.Element("defs")
+def get_hole_mask(board, defs):
     mask = etree.SubElement(defs, "mask", id="hole-mask")
     container = etree.SubElement(mask, "g")
 
+    board.SetVisibleAlls()
     bb = board.ComputeBoundingBox()
     bg = etree.SubElement(container, "rect", x="0", y="0", fill="white")
     bg.attrib["x"] = str(ki2svg(bb.GetX()))
@@ -540,7 +553,6 @@ def get_hole_mask(board):
                 el.attrib["points"] = points
                 el.attrib["transform"] = "translate({} {}) rotate({})".format(
                     pos.x, pos.y, -padOrientation / 10)
-    return defs
 
 def get_model_file(paths, lib, name, ref, remapping):
     """ Find model file in library considering component remapping """
@@ -568,43 +580,72 @@ def component_to_board_scale(svg):
     x, y, vw, vh = [float(x) for x in svg.attrib["viewBox"].split()]
     return width / vw, height / vh
 
-def component_from_library(lib, name, value, ref, pos, comp, highlight, silent, no_warn_back):
+def component_from_library(lib, name, value, ref, pos, usedComponents, comp,
+                           highlight, silent, no_warn_back):
     if not name:
         return
     if comp["filter"] is not None and ref not in comp["filter"]:
         return
-    f = get_model_file(comp["libraries"], lib, name, ref, comp["remapping"])
-    if not f:
-        if not silent:
-            if name[-5:] != '.back' or not no_warn_back:
-                print("Warning: component '{}' for footprint '{}' from library '{}' was not found".format(name, ref, lib))
-        if comp["placeholder"]:
-            etree.SubElement(comp["container"], "rect", x=str(ki2svg(pos[0] - mm2ki(0.5))), y=str(ki2svg(pos[1] - mm2ki(0.5))),
-                             width=str(ki2svg(mm2ki(1))), height=str(ki2svg(mm2ki(1))), style="fill:red;")
-        return
+
+    unique_name = f"{lib}__{name}"
+    if unique_name in usedComponents:
+        componentInfo = usedComponents[unique_name]
+        componentElement = etree.Element("use", attrib={"{http://www.w3.org/1999/xlink}href": "#" + componentInfo["id"]})
+    else:
+        f = get_model_file(comp["libraries"], lib, name, ref, comp["remapping"])
+        if not f:
+            if not silent:
+                if name[-5:] != '.back' or not no_warn_back:
+                    print("Warning: component '{}' for footprint '{}' from library '{}' was not found".format(name, ref, lib))
+            if comp["placeholder"]:
+                etree.SubElement(comp["container"], "rect", x=str(ki2svg(pos[0] - mm2ki(0.5))), y=str(ki2svg(pos[1] - mm2ki(0.5))),
+                                width=str(ki2svg(mm2ki(1))), height=str(ki2svg(mm2ki(1))), style="fill:red;")
+            return
+        xml_id = make_XML_identifier(unique_name)
+        componentElement = etree.Element("g", attrib={"id": xml_id})
+        svg_tree = read_svg_unique(f)
+        for x in extract_svg_content(svg_tree):
+            if x.tag in ["namedview", "metadata"]:
+                continue
+            componentElement.append(x)
+        origin_x = 0
+        origin_y = 0
+        origin = componentElement.find(".//*[@id='origin']")
+        if origin is not None:
+            origin_x, origin_y = element_position(origin, root=componentElement)
+            origin.getparent().remove(origin)
+        else:
+            print("Warning: component '{}' from library '{}' has no ORIGIN".format(name, lib))
+        svg_scale_x, svg_scale_y = component_to_board_scale(svg_tree)
+        componentInfo = {
+            "id": xml_id,
+            "origin_x": origin_x,
+            "origin_y": origin_y,
+            "scale_x": svg_scale_x,
+            "scale_y": svg_scale_y,
+            "width": svg_tree.attrib["width"],
+            "height": svg_tree.attrib["height"]
+        }
+        usedComponents[unique_name] = componentInfo
+
     comp["container"].append(etree.Comment("{}:{}".format(lib, name)))
     r = etree.SubElement(comp["container"], "g")
-    svg_tree = read_svg_unique(f)
-    for x in extract_svg_content(svg_tree):
-        r.append(x)
-    origin_x = 0
-    origin_y = 0
-    origin = r.find(".//*[@id='origin']")
-    if origin is not None:
-        origin_x, origin_y = element_position(origin, root=r)
-        origin.getparent().remove(origin)
-    else:
-        print("Warning: component '{}' from library '{}' has no ORIGIN".format(name, lib))
+    r.append(componentElement)
+    svg_scale_x = componentInfo["scale_x"]
+    svg_scale_y = componentInfo["scale_y"]
+    origin_x = componentInfo["origin_x"]
+    origin_y = componentInfo["origin_y"]
+    width = componentInfo["width"]
+    height = componentInfo["height"]
 
-    svg_scale_x, svg_scale_y = component_to_board_scale(svg_tree)
     r.attrib["transform"] = \
         f"translate({ki2svg(pos[0])} {ki2svg(pos[1])}) " + \
         f"scale({svg_scale_x} {svg_scale_y}) " + \
         f"rotate({-math.degrees(pos[2])}) " + \
         f"translate({-origin_x}, {-origin_y})"
     if ref in highlight["items"]:
-        w = ki2svg(to_kicad_basic_units(svg_tree.attrib["width"]))
-        h = ki2svg(to_kicad_basic_units(svg_tree.attrib["height"]))
+        w = ki2svg(to_kicad_basic_units(width))
+        h = ki2svg(to_kicad_basic_units(height))
         build_highlight(highlight, w, h, pos, (origin_x, origin_y), (svg_scale_x, svg_scale_y), ref)
 
 def build_highlight(preset, width, height, pos, origin, scale, ref):
@@ -746,6 +787,29 @@ def shrink_svg(svgfilepath, shrinkBorder):
     root.attrib["height"] = str(ki2mm(svg2ki(bbox[3] - bbox[2]))) + "mm"
     document.save(svgfilepath)
 
+def remove_empty_elems(tree):
+    """
+    Given SVG tree, remove empty groups and defs
+    """
+    for elem in tree:
+        remove_empty_elems(elem)
+    toDel = []
+    for elem in tree:
+        if elem.tag in ["g", "defs"] and len(elem.getchildren()) == 0:
+            toDel.append(elem)
+    for elem in toDel:
+        tree.remove(elem)
+
+def remove_inkscape_annotation(tree):
+    for elem in tree:
+        remove_inkscape_annotation(elem)
+    for key in tree.attrib.keys():
+        if "inkscape" in key:
+            tree.attrib.pop(key)
+    # Comments have callable tag...
+    if not callable(tree.tag):
+        objectify.deannotate(tree, cleanup_namespaces=True)
+
 def postprocess_svg(svgfilepath, shrinkBorder):
     if shrinkBorder is not None:
         shrink_svg(svgfilepath, shrinkBorder)
@@ -767,7 +831,7 @@ def main():
     parser.add_argument("-s", "--style", help="JSON file with board style")
     parser.add_argument("board", help=".kicad_pcb file to draw")
     parser.add_argument("output", help="destination for final SVG or PNG file")
-    parser.add_argument("-l", "--libs", help="coma separated list of libraries; use default, kicad-default or eagle-default for built-in libraries", default="default")
+    parser.add_argument("-l", "--libs", help="comma separated list of libraries; use default, kicad-default or eagle-default for built-in libraries", default="default")
     parser.add_argument("-p", "--placeholder", action="store_true",
                         help="show placeholder for missing components")
     parser.add_argument("-m", "--remap",
@@ -819,6 +883,7 @@ def main():
                                         remapping=remapping))
         sys.exit(0)
 
+    board.SetVisibleAlls()
     bb = board.ComputeBoundingBox()
     transform_string = ""
     # Let me briefly explain what's going on. KiCAD outputs SVG in user units,
@@ -837,6 +902,7 @@ def main():
             height=f"{ki2mm(bb.GetHeight())}mm",
             viewBox=f"{ki2svg(-bb.GetWidth() - bb.GetX())} {ki2svg(bb.GetY())} {ki2svg(bb.GetWidth())} {ki2svg(bb.GetHeight())}")
 
+    defs = etree.SubElement(document.getroot(), "defs")
     board_cont = etree.SubElement(document.getroot(), "g", transform=transform_string)
     if style["highlight-on-top"]:
         comp_cont = etree.SubElement(document.getroot(), "g", transform=transform_string)
@@ -864,17 +930,23 @@ def main():
         "padding": style["highlight-padding"]
     }
 
-    board_cont.append(get_board_substrate(board, style, not args.no_drillholes, args.back))
+    board_cont.append(get_board_substrate(board, style, defs, not args.no_drillholes, args.back))
     if args.vcuts:
-        board_cont.append(get_layers(board, style, [("vcut", [pcbnew.Cmts_User], process_board_substrate_layer)]))
+        board_cont.append(get_layers(board, style, defs, [("vcut", [pcbnew.Cmts_User], process_board_substrate_layer)]))
 
+    usedComponents = {}
     walk_components(board, args.back, lambda lib, name, val, ref, pos:
-        component_from_library(lib, name, val, ref, pos, components, highlight, args.silent, args.no_warn_back))
+        component_from_library(lib, name, val, ref, pos, usedComponents,
+                               components, highlight, args.silent, args.no_warn_back))
 
     # make another pass for search, and if found, render the back side of the component
     # the function will search for file with extension ".back.svg"
     walk_components(board, not args.back, lambda lib, name, val, ref, pos:
-        component_from_library(lib, name+".back", val, ref, pos, components, highlight, args.silent, args.no_warn_back))
+        component_from_library(lib, name+".back", val, ref, pos, usedComponents,
+                                components, highlight, args.silent, args.no_warn_back))
+
+    remove_empty_elems(document.getroot())
+    remove_inkscape_annotation(document.getroot())
 
     if args.output.endswith(".svg") or args.output.endswith(".SVG"):
         document.write(args.output)
@@ -883,9 +955,9 @@ def main():
         with tempfile.NamedTemporaryFile(suffix=".svg") as tmp_f:
             document.write(tmp_f)
             tmp_f.flush()
-            postprocess_svg(args.output, args.shrink)
+            postprocess_svg(tmp_f.name, args.shrink)
             tmp_f.flush()
             svg_to_bitmap(tmp_f.name, args.output, dpi=args.dpi)
 
 if __name__ == '__main__':
-    main()
\ No newline at end of file
+    main()
